
# This script runs on the remote server and implements a REPL.  Commands are
# received, executed, and results returned.

import os, sys, traceback, functools
from io import StringIO
from os.path import isdir, exists, abspath
import importlib.util

from .protocol import encode, read_msg, hash_file


class Message:
    def __init__(self, msgdict):
        self.__dict__.update(msgdict)


class PepperInternalError(Exception):
    """
    Used for "expected" exceptions which should not print a stack trace.
    Perhaps we should use SystemExit for this?
    """
    pass


def writemsg(msg):
    header, msgbytes = encode(msg)
    sys.__stdout__.buffer.write(header)
    sys.__stdout__.buffer.write(msgbytes)
    sys.__stdout__.buffer.flush()


def main():
    """
    The REPL that reads commands, executes them, and returns results.  This
    loops until "quit" is received.
    """

    unbuffered = os.fdopen(sys.stdin.fileno(), 'rb', buffering=0)

    try:
        while 1:
            msg = read_msg(functools.partial(unbuffered.read, 50 * 1024))
            msg = Message(msg)

            if msg.msgtype == 'quit':
                break

            success = False
            try:
                func = globals().get('_handle_%s' % msg.msgtype)

                if not func:
                    raise PepperInternalError('Unhandled msgtype %r' % msg.msgtype)

                response = func(msg)
                success = True

            except PepperInternalError as ex:
                response = dict(msgtype='error', error=str(ex))
            except:
                response = dict(msgtype='error', traceback=traceback.format_exc())

            writemsg(response)
            if not success:
                break

    except:
        writemsg(dict(msgtype='error', traceback=traceback.format_exc()))


def get_lib_root():
    """
    Returns the directory where source is installed.  The directory is in the
    path so packages can be installed under it.
    """
    # We happen to know that the client pushed the temporary directory as the
    # first element of the path.
    assert isdir(sys.path[0])
    return sys.path[0]


def _handle_script(msg):
    # The client has already copied the script via SFTP and passes us the path to it.
    #
    # * modulename: The name of the module.
    # * filename: The path to it on this box.

    spec = importlib.util.spec_from_file_location(msg.modulename, msg.filename)
    module = importlib.util.module_from_spec(spec)

    sys.stdout = StringIO()
    sys.stderr = sys.stdout
    try:
        spec.loader.exec_module(module)
        sys.modules[msg.modulename] = module

        return dict(
            msgtype='script',
            modulename=msg.modulename,
            stdout=sys.stdout.getvalue()
        )
    except:
        return dict(msgtype='error', traceback=traceback.format_exc(),
                    stdout=sys.stdout.getvalue())
    finally:
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__


def _handle_call(msg):
    module = sys.modules[msg.mname]
    try:
        func = getattr(module, msg.fname)
    except AttributeError as ex:
        return dict(msgtype='error', error='no attribute',
                    traceback=traceback.format_exc())

    sys.stdout = StringIO()
    sys.stderr = sys.stdout
    try:
        result = func(*msg.args, **msg.kwargs)
        return dict(msgtype='callret',
                    result=result,
                    stdout=sys.stdout.getvalue())
    except:
        return dict(msgtype='error', traceback=traceback.format_exc(),
                    stdout=sys.stdout.getvalue())
    finally:
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__


def _handle_stat(msg):

    # 'exists isdir isfile size hash'

    results = {}

    if isdir(msg.filename):
        results['exists'] = True
        results['isdir']  = True
        results['isfile'] = False
    elif exists(msg.filename):
        results['exists'] = True
        results['isdir']  = False
        results['isfile'] = True
        results['size'] = os.path.getsize(msg.filename)
        results['hash'] = hash_file(msg.filename)

    results['msgtype'] = 'stat'
    results['filename'] = msg.filename
    results['fqn'] = abspath(msg.filename)

    return results




import os, zipfile, fnmatch, glob
import subprocess
from os.path import isabs, join, relpath
from contextlib import contextmanager


def shell(cmd, check=True):
    """
    Execute cmd.  Raises an exception if the command does not return 0
    """
    proc = subprocess.run(cmd, shell=True, check=False, stderr=subprocess.STDOUT,
                          stdout=subprocess.PIPE)
    output = proc.stdout.decode('utf-8').strip()
    if output:
        print(output)
    if check:
        proc.check_returncode()


def ensurelist(paramname, val, comment='#'):
    """
    Returns `val` as a list of strings.

    If `val` is already a list of strings, it is returned as-is.  If a string
    is passed, it is split into lines and blink and (optionally) comment lines
    are removed.

    paramname
      The parameter name to put into error messages.

    comment
      The single-line comment identifier.  Any lines starting with this are
      removed.  Note that end-of-line comments are *not* removed.
    """

    if isinstance(val, list):
        for line in val:
            if not isinstance(line, str):
                typename = type(line).__qualname__
                raise TypeError(f"{paramname} should be a list of strings.  Found: {typename}")
        return val

    if isinstance(val, str):
        lines = (line.strip() for line in val.strip().split('\n'))
        lines = (line for line in lines if line)
        if comment:
            lines = (line for line in lines if not line.startswith())
        return lines

    raise TypeError(f'{paramname} must be a list of strings or a string')


def archive(filename, *, include, root=None, exclude_files=None,
            nocompression=['*.png', '*.jpg']):
    """
    Creates a zip file.

    root
      An optional directory used as the parent of all relative paths in other
      parameters.  If not provided, the current working directory is used.

    include
      A list of glob patterns to be added.  For convenience a string can be
      passed which will be split on newlines, allowing a triple-quoted string
      to be used.

      Each pattern is treated as a recursive pattern.  To disable, end the
      pattern with '/*'0

    exclude_files
      A list of shell patterns to exclude, such as '*.log'.  The fnmatch
      package is used to compare.
    """
    if not root:
        root = os.getcwd()

    if isinstance(include, str):
        lines = (line.strip() for line in include.split('\n'))
        nonblank = (line for line in lines
                    if line and not line.startswith('#'))
        include = list(nonblank)

    archive = zipfile.ZipFile(filename, 'w', compression=zipfile.ZIP_DEFLATED)

    def _matches_pattern(fn, patterns):
        if not patterns:
            return False
        for pattern in patterns:
            if fnmatch.fnmatch(fn, pattern):
                return True
        return False

    for spec in include:
        if isabs(spec):
            raise ValueError(
                f'Include specifications cannot be absolute.  ({spec!r})'
            )
        spec = join(root, spec)
        recursive = not spec.endswith('/*')
        for fn in glob.iglob(spec, recursive=recursive):

            if _matches_pattern(fn, exclude_files):
                continue

            compression = None
            if _matches_pattern(fn, nocompression):
                compression = zipfile.ZIP_STORED

            arcname = relpath(fn, root)
            archive.write(fn, arcname, compress_type=compression)

    return filename


@contextmanager
def chdir(dir):
    prev = os.getcwd()
    try:
        os.chdir(dir)
        yield
    finally:
        os.chdir(prev)



#!/usr/bin/env python

from distutils.core import setup

VERSION = '1.2.0'

setup(
    name='pepperssh',
    description='Simple SSH based remote execution',
    version=VERSION,
    package_dir={'': 'src'},
    packages=['pepperssh'],
    author='Michael Kleehammer',
    author_email='michael@kleehammer.com',
    url='https://gitlab.com/mkleehammer/pepperssh'
)
